#!/usr/bin/env python

from distutils.core import setup

setup(name='Tachyon',
      version='1.0.1',
      description='Fast Sum Product Network Library',
      author='Agastya Kalra',
      author_email='a6kalra@uwaterloo.ca',
      url='https://github.com/KalraA/Tachyon',
      packages=['tachyon'],
      py_modules=['numpy', 'tensorflow']
     )
